              /* Functions for point and line insertion in a CDT */
#include "dcel.h"
#include "circular.h"
#include "relation.h"
#include "attribut.h"
#include "geometry.h"
#include "euler.h"
/* #include "debug.h" */  /* Header File non disponibile */

#ifndef TRUE
typedef int bool;
#define TRUE 1
#define FALSE 0
#endif

bool INSERT_POINT(T,x0,y0)
sptr T; double x0,y0;
/* Insert a new vertex (x0,y0) into triangulation T
   and update it as a consequence */
{
 fptr f,t1,tnew;eptr e,e1,e2,enew;vptr v,vnew;

 f=ENCLOSING_FACE(T,x0,y0);
 if (IS_INFINITY_FACE(T,f)) return(FALSE); /*point is outside the convex hull*/
 e1=FIRST_BOUNDING_EDGE(T,f); /* f is a triangle */
 do
 {
  MARK(e1);
  if (IS_A_CONSTRAINT(T,e1) || IS_BOUNDARY(T,e1))
   e1=NEXT_CLOCK_EDGE(T,f,e1);
  else
  {
   t1=OTHER_FACE(T,f,e1); /* t1 is always a triangle */
   if (IS_IN_CIRCUMCIRCLE(T,t1,x0,y0))
   {
    e2=NEXT_CLOCK_EDGE(T,t1,e1);
    if (!CANCEL_MARK((iptr)e1))
    {
     printf("\nCM\n");
     exit(1);
    }
    KILL_EDGE_FACE(T,&e1,&t1);
    e1=e2;
   }
   else
    e1=NEXT_CLOCK_EDGE(T,f,e1);
  }
 } while (!IS_MARKED(e1));
 e=e1;
 e2=NEXT_CLOCK_EDGE(T,f,e1);
 v=COMMON_VERTEX(T,e1,e2);
 MAKE_EDGE_VERTEX(T,v,x0,y0,&vnew,&enew);
 while (e2!=e)
 {
  e1=e2;
  e2=NEXT_CLOCK_EDGE(T,f,e1);
  v=COMMON_VERTEX(T,e1,e2);
      /* faces tnew and f lie by convention left and right
                                         respectively wrt vector v-vnew */
  MAKE_EDGE_FACE(T,f,&tnew,vnew,v,&enew);
 }
 RESET_MARK();
 return(TRUE);
}

bool INFLUENCE_POLYGONS(T,vf,vl,f1,f2)
sptr T; vptr vf,vl;fptr *f1,*f2;
/* compute the influence polygons f1 and f2 of constraint edge connecting
   vertices vf and vl in triangulation T. Delete from T the edges crossed
   by the constraint edge */
{
 lptr Lt; fptr t1; vptr v1,v2; eptr e,e1; double xf,yf,xl,yl,x1,y1,x2,y2;

 if (!(IS_A_VERTEX(T,vf) && IS_A_VERTEX(T,vl))) return(FALSE);
 Lt=VERTEX_FACE(T,vf);
 COORDINATES(vf,&xf,&yf);
 COORDINATES(vl,&xl,&yl);
 do          /* find edge intersecting l nearest to vf */
 {
  (*f1)=(fptr)FIRST(Lt);
  CLOCKWISE(&Lt);
  e=NEXT_CLOCK_EDGE(T,*f1,CLOCK_INCIDENT_EDGE(T,*f1,vf));
  EDGE_VERTEX(T,e,&v1,&v2);
  COORDINATES(v1,&x1,&y1);
  COORDINATES(v2,&x2,&y2);
 } while (!SEGMENT_INTERSECT(xf,yf,xl,yl,x1,y1,x2,y2));
 FREELIST(&Lt);  /* aggiunta dai "tesinisti": la lista non era deallocata... */
 do
 {
  t1=OTHER_FACE(T,*f1,e);
  e1=NEXT_CLOCK_EDGE(T,t1,e);
  if (IS_A_CONSTRAINT(T,e)) return(FALSE);
  KILL_EDGE_FACE(T,&e,&t1);
  if (!IS_INCIDENT(T,e1,vl))
  {
   EDGE_VERTEX(T,e1,&v1,&v2);
   COORDINATES(v1,&x1,&y1);
   COORDINATES(v2,&x2,&y2);
   if (SEGMENT_INTERSECT(xf,yf,xl,yl,x1,y1,x2,y2))
    e=e1;
   else
    e=NEXT_CLOCK_EDGE(T,*f1,e1);
  }
 } while (!IS_INCIDENT(T,e1,vl));
 MAKE_EDGE_FACE(T,*f1,f2,vf,vl,&e);
 SET_CONSTRAINT_EDGE(T,e);
 return(TRUE);
}

static void TEP(T,f,l,fl)
sptr T; fptr f; eptr l; lptr *fl;
{
 vptr v,vf,vl;lptr Lv; eptr ef,el,enew1,enew2;fptr f1,f2; double xf,yf,xl,yl;

 if (!IS_A_TRIANGLE(T,f))
 {
  ef=NEXT_CLOCK_EDGE(T,f,l);   /* find vertex of f nearest to l */
  vf=COMMON_VERTEX(T,l,ef);
  vl=OTHER_VERTEX(T,vf,l);
  COORDINATES(vf,&xf,&yf);
  COORDINATES(vl,&xl,&yl);
  Lv=FACE_VERTEX(T,f);
  v=NEAREST_VERTEX(Lv,xf,yf,xl,yl);
  FREELIST(&Lv); /* aggiunta dai "tesinisti": la lista non era deallocata... */
  el=CLOCK_INCIDENT_EDGE(T,f,v);
  if (!IS_INCIDENT(T,ef,v))
  {
   MAKE_EDGE_FACE(T,f,&f1,vf,v,&enew1);
   INSERT_CLOCK(fl,(iptr)f1);
   F_MARK(f1);
   TEP(T,f1,enew1,fl);
  }
  if (!IS_INCIDENT(T,el,vl))
  {
   MAKE_EDGE_FACE(T,f,&f2,v,vl,&enew2);
   INSERT_CLOCK(fl,(iptr)f2);
   F_MARK(f2);
   TEP(T,f2,enew2,fl);
  }
 }
}

bool TRIANGULATE_ESS_POLYGON(T,f,l,fl)
sptr T; fptr f; eptr l; lptr *fl;
/* triangulate polygon f of tessellation T - f is edge star-shaped wrt
   edge l - return list fl of new triangles inserted into T */
{
 if (!IS_BOUNDING(T,l,f)) return(FALSE);
 INIT(fl);
 F_MARK(f);
 INSERT_CLOCK(fl,(iptr)f);
 TEP(T,f,l,fl);
 return(TRUE);
}

void OPTIMIZE(T,fl)
sptr T; lptr *fl;
/* Optimize the poligonal region of triangulation T formed by the
   triangles of list fl. Procedure uses edge swapping based on
   circumcircle criterion. Triangles of fl are (the only triangles in tr)
   marked at the beginning */
{
 lptr fl1;vptr v; eptr e,e1;fptr curt,t1,lastt=NULL; double xv,yv;
 int i; bool ifl,iflag;

 do             /* repeat optimization until there are no swapping */
 {
  iflag=FALSE;
  fl1=MARK_POSITION(*fl);
  do            /* for every triangle of fl */
  {
   ifl=FALSE;
   curt=(fptr)FIRST(*fl);
   if (F_MARKED(curt))   /* if curt has to be checked */
   {
    e=FIRST_BOUNDING_EDGE(T,curt);
    e1=NEXT_CLOCK_EDGE(T,curt,e);
    for (i=1;i<=3;i++)  /* for every neighbor of curt */
    {
     t1=OTHER_FACE(T,curt,e);
     if (F_MARKED(t1) && (t1!=lastt) && !IS_A_CONSTRAINT(T,e))
     {
      v=OPPOSITE_VERTEX(T,t1,e);
      COORDINATES(v,&xv,&yv);
      if (IS_IN_CIRCUMCIRCLE(T,curt,xv,yv))
      {
       CANCEL(fl,&t1);
       SWAP_EDGE(T,&curt,&t1,&e);
       INSERT_CLOCK(fl,(iptr)t1);
       iflag=TRUE;
       ifl=TRUE;
       lastt=t1;
      }
     }
     e=e1;
     e1=NEXT_CLOCK_EDGE(T,curt,e1);
    }
   }
   if (!ifl) 
    F_UNMARK(curt); /* if there has been no swapping
                       t has not to be checked any more */
   CLOCKWISE(fl);
  } while (!SAME_POSITION(*fl,fl1));
 } while (iflag);
 F_RESET_MARK();
}
