           /* Geometric functions for 2D tesselations based on DCEL */
#include "dcel.h"
#include "circular.h"
#include "relation.h"

#include <math.h>

#define odd(n) (n)%2

/* #include "debug.h" */  /* Header File non disponibile */
/* #include "input.h" */  /* Header File non disponibile */
#include <stdio.h>
extern double yhigh;

double DISTANCE(x0,y0,x1,y1)
double x0,y0,x1,y1;
/* return the euclidean distance between points v and v1 */
{
 x0 -= x1;
 y0 -= y1;
 return(sqrt(x0*x0+y0*y0));
}

double LINE_DISTANCE(p,x1,y1,x2,y2)
vptr p; double x1,y1,x2,y2;
{
 double x0,y0,dx,dy;

 x0=p->x;
 y0=p->y;
 if (x1==x2)
  return(fabs(x1-x0));
 else if (y1==y2)
  return(fabs(y1-y0));
 else
 {
  dx=x2-x1;
  dy=y2-y1;
  return(fabs(dy*x0-dx*y0+x2*y1-x1*y2)/sqrt(dx*dx+dy*dy));
 }
}
           
bool SEGMENT_INTERSECT(x1,y1,x2,y2,x3,y3,x4,y4)
double x1,y1,x2,y2,x3,y3,x4,y4;
/* return TRUE iff segment from (x1,y1) to (x2,y2) intersects
   segment from (x3,y3) to (x4,y4) */
{
 bool i12, i21;

 /* check if segment (x3,x4) intersects line thru x1 and x2 */
 if (x1==x2)
  i12=(((x3<=x1)&&(x4>=x1))||((x3>=x1)&&(x4<=x1)));
 else if (y1==y2)
  i12=(((y3<=y1)&&(y4>=y1))||((y3>=y1)&&(y4<=y1)));
 else
  i12=(
       (
        ((x3-x1)*(y2-y1)>=(y3-y1)*(x2-x1)) &&
        ((x4-x1)*(y2-y1)<=(y4-y1)*(x2-x1))
       ) ||
       (
        ((x3-x1)*(y2-y1)<=(y3-y1)*(x2-x1)) &&
        ((x4-x1)*(y2-y1)>=(y4-y1)*(x2-x1))
       )
      );
 /* check if segment (x1,x2) intersects line thru x3 and x4 */
 if (x3==x4)
  i21=((x1<=x3)&&(x2>=x3))||((x1>=x3)&&(x2<=x3));
 else if (y3==y4)
  i21=((y1<=y3)&&(y2>=y3))||((y1>=y3)&&(y2<=y3));
 else
  i21=(
       (
        ((x1-x3)*(y4-y3)>=(y1-y3)*(x4-x3)) &&
        ((x2-x3)*(y4-y3)<=(y2-y3)*(x4-x3))
       ) ||
       (
        ((x1-x3)*(y4-y3)<=(y1-y3)*(x4-x3)) &&
        ((x2-x3)*(y4-y3)>=(y2-y3)*(x4-x3))
       )
      );
 return(i12 && i21);
}

bool LINE_INTERSECTION(x1,y1,x2,y2,x3,y3,x4,y4,xi,yi)
double x1,y1,x2,y2,x3,y3,x4,y4,*xi,*yi;
/* return in (xi,yi) the intersection between line thru (x1,y1) and (x2,y2)
   line thru (x3,y3) and (x4,y4). Return TRUE iff the two lines are parallel */
{
 double m,n,p,q; bool parallel=FALSE;

 if (x1==x2)
  if (x3==x4)
   parallel=TRUE;
  else
  {
   *xi=x1;
   *yi=(y4-y3)*(*xi-x3)/(x4-x3)+y3;
  }
 else if (x3==x4)
 {
  *xi=x3;
  *yi=(y2-y1)*(*xi-x1)/(x2-x1)+y1;
 }
 else
 {
  m=(y2-y1)/(x2-x1);
  p=y1-x1*m;
  n=(y4-y3)/(x4-x3);
  q=y3-x3*n;
  if (m==n)
   parallel=TRUE;
  else
  {
   *xi=(q-p)/(m-n);
   *yi=(*xi)*m+p;
  }
 }
 return(parallel);
}

bool CIRCUMCIRCLE(s,t,x0,y0,r)
sptr s; fptr t; double *x0,*y0,*r;
{
 double xv10,yv10,xv11,yv11,xv20,yv20,xv21,yv21,xm1,ym1,xm2,ym2;
 lptr el; eptr e1,e2;

 if (IS_A_TRIANGLE(s,t))
 {
  if (t->rc==0)
  {
   el=FACE_EDGE(s,t);
   e1=(eptr)FIRST(el);
   xv10=e1->vertices[0]->x;
   yv10=e1->vertices[0]->y;
   xv11=e1->vertices[1]->x;
   yv11=e1->vertices[1]->y;
   xm1=(xv10+xv11)/2;    /* compute middle point of an edge of t */
   ym1=(yv10+yv11)/2;
   CLOCKWISE(&el);
   e2=(eptr)FIRST(el);
   xv20=e2->vertices[0]->x;
   yv20=e2->vertices[0]->y;
   xv21=e2->vertices[1]->x;
   yv21=e2->vertices[1]->y;
   xm2=(xv20+xv21)/2;    /* compute middle point of another edge of t */
   ym2=(yv20+yv21)/2;
   LINE_INTERSECTION(xm1+ym1-yv10,ym1+xv10-xm1,xm1+yv10-ym1,ym1+xm1-xv10,
                     xm2+ym2-yv20,ym2+xv20-xm2,xm2+yv20-ym2,ym2+xm2-xv20,
                     &t->xc,&t->yc); /*compute circumcenter coordinates*/
   t->rc=DISTANCE(t->xc,t->yc,xv10,yv10);  /* compute circumcircle radius */
   FREELIST(&el);
  };
  *x0=t->xc;*y0=t->yc;*r=t->rc;
 }
 else
 {
  *x0=0;*y0=0;*r=0;
 }
}
            
bool IS_IN_CIRCUMCIRCLE(s,t,x0,y0)
sptr s; fptr t; double x0,y0;
{
 double xc,yc,r;

 CIRCUMCIRCLE(s,t,&xc,&yc,&r);
 return(DISTANCE(x0,y0,xc,yc)<r);
}

static bool greater(e,xp,yp,tgp,center,x0,y0)
eptr e; double xp,yp,tgp,x0,y0; vptr center;
/* returns TRUE if the angle between the vertical line from center
   and e is greater than the angle between the same vertical line
   and the segment for center and (xp,yp) */
{
 double xe,ye,tge;vptr v;

 v=e->vertices[center==(e->vertices[0])];
 xe=v->x-x0;ye=v->y-y0;
 if (xe==0)
  if (ye>0)                 /* 0 degrees for e*/
   return((xp==0)&&(yp>0)); /* equal angles */
  else                        /* 180 degrees for e */
   return(xp<=0);
 else
  if (xp==0)
   if (yp>0)                 /* 0 degrees for p */
    return(TRUE);
   else                        /* 180 degrees for p */
    return(xe>0);
  else
  {                          /* general case */
   tge=ye/xe;
   return((((xp*xe)>0)&&(tgp<=tge))||((xp<0)&&(xe>0)));
  }
}

bool SECTOR(center,l,p,ecck,eck)
vptr center,p; lptr l; eptr *ecck,*eck;
/* l is a list of edges incident into vertex center. p is a point
   in the plane. ecck and eck (output parameters) are the edges
   bounding the sector of plane containing p, in counterclockwise
   and clockwise position wrt p respectively */
{
 double x0,y0,xp,yp,tgp,x1,y1,tg1; eptr emin,emax; vptr vv;

 if (EMPTY(l)) return(FALSE);
 x0=center->x;y0=center->y;
 xp=p->x-x0;yp=p->y-y0;
 if (xp)
  tgp=yp/xp;
  /* find max and min edge around center according to the order
     used by Function greater */
 do
 {
  emin=(eptr)FIRST(l);
  vv=emin->vertices[center==(emin->vertices[0])];
  x1=vv->x-x0;
  y1=vv->y-y0;
  if (x1!=0)
   tg1=y1/x1;
  CLOCKWISE(&l);
  emax=(eptr)FIRST(l);
 } while (!greater(emax,x1,y1,tg1,center,x0,y0));
 if (greater(emin,xp,yp,tgp,center,x0,y0)||
     (!greater(emax,xp,yp,tgp,center,x0,y0)))
 /* p is in sector between emax and emin */
 {
  *ecck=emin;
  *eck=emax;
 }
 else       /* search sector */
 {
  while (greater(FIRST(l),xp,yp,tgp,center,x0,y0))
   CLOCKWISE(&l);
  *eck=(eptr)FIRST(l);
  COUNTERCLOCK(&l);
  *ecck=(eptr)FIRST(l);
 }
 return(TRUE);
}

static int check_cusp(s,l,x0,y0)
sptr s; lptr l; double x0,y0;
/* check intersections between r and vertices of edges in list l. 
   Return the number of simple intersections (no tangent) */
{
 lptr l0; vptr v1,v2,v3; eptr e1,e2; double x1,x2,x3,y2; int count=0;

 l0=MARK_POSITION(l);
 e2=(eptr)FIRST(l);
 do
 {
  e1=e2;
  CLOCKWISE(&l);
  e2=(eptr)FIRST(l);
  if (e1!=e2)
   v2=COMMON_VERTEX(s,e1,e2);
  else
   v2=e1->vertices[IS_DANGLING_VERTEX(s,e1->vertices[1])];
  x2=v2->x;y2=v2->y;
  if ((x0==x2)&&(y0<y2))
  {
   v1=OTHER_VERTEX(s,v2,e1);
   v3=OTHER_VERTEX(s,v2,e2);
   x1=v1->x;
   x3=v3->x;
   if (((x1<=x2)&&(x2<=x3))||((x1>=x2)&&(x2>=x3)))
    count++;
  }
 } while (!SAME_POSITION(l,l0));
 return(count);
}
                   
static int intersect_y(e,x0,y0)
eptr e; double x0,y0;
/* check intersections between (open) edge e and r.
   Return 0==no intersection or tangent, 1==intersection */
{
 double x1,y1,x2,y2;

 x1=e->vertices[1]->x;
 y1=e->vertices[1]->y;
 x2=e->vertices[0]->x;
 y2=e->vertices[0]->y;
 if (((x1<x0)&&(x2>x0))||((x1>x0)&&(x2<x0)))
  if ((y2-y1)*(x0-x1)/(x2-x1)+y1 > y0)
   return(1);
  else 
   return(0);
 else
  return(0);
}

bool ENCLOSES(s,f,x0,y0)
sptr s; fptr f; double x0,y0;
/* s is a tessellation, f a face of s, (x0,y0) is a point in the plane.
   Return TRUE if f encloses the point. Algorithm (based on Jordan curve
   theorem) uses vertical half-line r: x=x0,y>y0. Algorithm checks first
   special cases where r intersects some vp vertex of f, then checks
   intersections on open edges of f */
{
 int ints;lptr l,l0;bool ext;

 l=FACE_EDGE(s,f);
 if (EMPTY(l))    /* only infinity face */
  return(TRUE);
 else
 {
  ints=check_cusp(s,l,x0,y0);
  l0=MARK_POSITION(l);
  do
  {
   ints+=intersect_y((eptr)FIRST(l),x0,y0);
   CLOCKWISE(&l);
  } while (!SAME_POSITION(l,l0));
  ext=(f==INFINITY_FACE(s));
  FREELIST(&l);
  return((odd(ints)&&(!ext))||(ext && (!odd(ints))));
 }
}

fptr ENCLOSING_FACE(s,x0,y0)
sptr s; double x0,y0;
/* s is a tessellation of the plane, (x0,y0) is a point in
   the plane. Return the face of s that encloses the point p.
   Brute-force full-scan algorithm */
{
 fptr f;

 f=FIRST_FACE(s);
 while (!ENCLOSES(s,f,x0,y0))
  f=NEXT_FACE(s,f);
 return(f);
}

static bool same_point(v,xp,yp)
vptr v; double xp,yp;
{
 double x,y;

 COORDINATES(v,&x,&y);
 return((x==xp)&&(y==yp));
}

vptr NEAREST_VERTEX(vl,x1,y1,x2,y2)
lptr vl; double x1,y1,x2,y2;
/* return the vertex in list vl that is closest to line thru points
   (x1,y1) and (x2,y2) - exclude points (x1,y1) and (x2,y2) from list */
{
 vptr v,vmin;double d,dmin;lptr l1;

 v=(vptr)FIRST(vl);
 while (same_point(v,x1,y1) || same_point(v,x2,y2))
 {
  CLOCKWISE(&vl);
  v=(vptr)FIRST(vl);
 }
 vmin=v;
 dmin=LINE_DISTANCE(v,x1,y1,x2,y2);
 l1=MARK_POSITION(vl);
 CLOCKWISE(&vl);
 do
 {
  v=(vptr)FIRST(vl);
  if (!(same_point(v,x1,y1) || same_point(v,x2,y2)))
  {
   d=LINE_DISTANCE(v,x1,y1,x2,y2);
   if (d<dmin)
   {
    dmin=d;
    vmin=v;
   }
  }
  CLOCKWISE(&vl);
 } while (!SAME_POSITION(vl,l1));
 return(vmin);
}

bool IS_ON_CONSTRAINT(T,vx,vy)
sptr T; double vx,vy;
/* verifica se un punto che si tenta di immettere, di coordinate (vx,vy),
   giaccia su un vincolo esistente della tassellazione */
{
 eptr ce; double x1,y1,x2,y2,dx,dy;

 ce = T->elist;
 while (ce != NULL) {
  if (!IS_A_CONSTRAINT(T,ce))
    ce = ce->next;
  else {
   x1 = ce->vertices[0]->x;
   y1 = ce->vertices[0]->y;
   x2 = ce->vertices[1]->x;
   y2 = ce->vertices[1]->y;
   if ((x1==x2) && (x1==vx))
    if (((y1 <= vy) && (vy <= y2)) || ((y2 <= vy) && (vy <= y1)))
     return(TRUE);
    else
     ce = ce->next;
   else if ((y1==y2) && (y1==vy))
    if (((x1 <= vx) && (vx <= x2)) || ((x2 <= vx) && (vx <= x1)))
     return(TRUE);
    else
     ce = ce->next;
   else {
    dx=x2-x1;
    dy=y2-y1;
    if ((fabs(dy*vx-dx*vy+x2*y1-x1*y2)/sqrt(dx*dx+dy*dy)) == 0)
      if (((x1 <= vx) && (vx <= x2)) || ((x2 <= vx) && (vx <= x1)))
        return(TRUE);
      else
       ce = ce->next;
    else
     ce = ce->next;
   }
  }
 };
 return(FALSE);
}

bool IS_ON_VERTEX(S,v1,v2)
sptr S; vptr v1,v2;
/* verifica se un vincolo che si tenta di immettere, di estremi (v1,v2),
   passi per un vertice esistente della tassellazione (non suo estremo) */
{
 double m,q; bool vertical = FALSE; vptr cv;
 const double EPS = 1.0E-12;   /* Valore di EPSilon usato per arrotondamenti */

 if (v1->y == v2->y) {
   m = 0.0;
   q = v1->y;
 }
 else if (v1->x == v2->x) {
  vertical = TRUE;
  q = v1->x;
 }
 else {
  m = (v2->y - v1->y) / (v2->x - v1->x);
  q = v1->y - m * v1->x;
 };
 cv = S->vlist;
 while (cv != NULL) {
   if (vertical)
     if ((cv->x == q) && (((cv->y > v1->y) && (cv->y < v2->y)) ||
                          ((cv->y > v2->y) && (cv->y < v1->y))))
       return(TRUE);
     else
       cv = cv->next;
   else if ((fabs(cv->y - (m * cv->x + q)) < EPS) &&
            (((cv->x > v1->x) && (cv->x < v2->x)) ||
             ((cv->x > v2->x) && (cv->x < v1->x))))
     return(TRUE);
   else
     cv = cv->next;
 }
 return(FALSE);
}
