            /* Access primitives for 2D tesselations based on DCEL */
#include "dcel.h"
#include "circular.h"
/* #include "debug.h" */  /* Header File non disponibile */

/* Query primitive functions */

bool IS_AN_EDGE(S,E)
sptr S;eptr E;
{
 if (E==NULL) return(FALSE);
 else return(S==E->structure);
}

bool IS_A_VERTEX(S,V)
sptr S;vptr V;
{
 if (S==NULL || V==NULL) return(FALSE);
 else
  if (V->incidentedge==NULL) return(S->vlist==V);
  else return(V->incidentedge->structure==S);
}

bool IS_A_FACE(S,F)
sptr S; fptr F;
{
 if (S==NULL || F==NULL) return(FALSE);
 else
  if (F->boundingedge==NULL) return(S->infinity==F);
  else return(F->boundingedge->structure==S);
}

bool IS_A_TRIANGLE(S,F)
sptr S; fptr F;
{
 eptr e1, e2, e3, e4;

 if (IS_A_FACE(S,F))
 {
  e1=F->boundingedge;
  e2=e1->edges[F==e1->faces[1]];
  e3=e2->edges[F==e2->faces[1]];
  e4=e3->edges[F==e3->faces[1]];
  return((e1!=e2)&&(e2!=e3)&&(e3!=e1)&&(e4==e1));
 }
 else return(FALSE);
}

bool IS_INCIDENT(S,E,V)
sptr S; eptr E; vptr V;
{
 if (IS_AN_EDGE(S,E))
  return(E->vertices[0]==V || E->vertices[1]==V);
 else return(FALSE);
}

bool IS_BOUNDING(S,E,F)
sptr S; eptr E; fptr F;
{
 if (IS_AN_EDGE(S,E) && IS_A_FACE(S,F))
  return(F==E->faces[0] || F==E->faces[1]);
 else return(FALSE);
}

bool IS_BOUNDARY(S,E)
sptr S; eptr E;
{
 if (IS_AN_EDGE(S,E))
  return(E->faces[0]==S->infinity || E->faces[1]==S->infinity);
 else  return(FALSE);
}

bool IS_ISOLATED(S,V)
sptr S; vptr V;
{
 if (IS_A_VERTEX(S,V)) return(V->incidentedge==NULL);
 else return(FALSE);
}

bool IS_DANGLING_VERTEX(S,V)
sptr S; vptr V;
{
 eptr e;

 if (IS_A_VERTEX(S,V))
  if (IS_ISOLATED(S,V)) return(TRUE);
  else
  {
   e=V->incidentedge;
   return(e->edges[V==e->vertices[1]]==e);
  }
 else return(FALSE);
}

bool IS_DANGLING_EDGE(S,E)
sptr S; eptr E;
{
 if (IS_AN_EDGE(S,E))
  return(E->edges[0]==E || E->edges[1]==E);
 else return(FALSE);
}

bool IS_INFINITY_FACE(S,F)
sptr S; fptr F;
{
 return(S->infinity==F);
}


/* Structure scan functions */

bool COORDINATES(v,x0,y0)
vptr v; double *x0, *y0;
{
 if (v==NULL) return(FALSE);
 else
 {
  *x0=v->x;
  *y0=v->y;
  return(TRUE);
 }
}

vptr FIRST_VERTEX(s)
sptr s;
{
 if (s==NULL) return(NULL);
 else return(s->vlist);
}
 
vptr NEXT_VERTEX(s,v)
sptr s; vptr v;
{
 if (IS_A_VERTEX(s,v)) return(v->next);
 else return(NULL);
}

vptr PREV_VERTEX(s,v)
sptr s; vptr v;
{
 if (IS_A_VERTEX(s,v)) return(v->prev);
 else return(NULL);
}

eptr FIRST_EDGE(s)
sptr s;
{
 if (s==NULL) return(NULL);
 else return(s->elist);
}
 
eptr NEXT_EDGE(s,e)
sptr s; eptr e;
{
 if (IS_AN_EDGE(s,e)) return(e->next);
 else return(NULL);
}

eptr PREV_EDGE(s,e)
sptr s; eptr e;
{
 if (IS_AN_EDGE(s,e)) return(e->prev);
 else return(NULL);
}

fptr FIRST_FACE(s)
sptr s;
{
 if (s==NULL) return(NULL);
 else return(s->flist);
}
 
fptr NEXT_FACE(s,f)
sptr s; fptr f;
{
 if (IS_A_FACE(s,f)) return(f->next);
 else return(NULL);
}

fptr PREV_FACE(s,f)
sptr s; fptr f;
{
 if (IS_A_FACE(s,f)) return(f->prev);
 else return(NULL);
}

fptr INFINITY_FACE(s)
sptr s;
{
 if (s==NULL) return(NULL);
 else return(s->infinity);
}


/* Relation primitive functions */

static void scan_chain(e1,v1,e,v,out0,out1)
eptr e,e1,*out0,*out1;vptr v,v1;
{
 *out0=e1;
 do
 {
  v1=e1->vertices[v1==e1->vertices[0]];
  e1=e1->edges[v1==e1->vertices[1]];
 } while (e1->vertices[v1==e1->vertices[0]]!=v ||
          e1->edges[v1==e1->vertices[0]]!=e);
 *out1=e1;
}

bool EDGE_EDGE(s,e,e00,e01,e10,e11)
sptr s; eptr e,*e00,*e01,*e10,*e11;
{
 vptr v1;

 if (! IS_AN_EDGE(s,e)) 
 {
  *e00=NULL;
  *e01=NULL;
  *e10=NULL;
  *e11=NULL;
  return(FALSE);
 }
 else
 {
  if (e->faces[0]==e->faces[1])      /* dangling chain */
  {
   v1=e->vertices[0];
   if (IS_DANGLING_VERTEX(s,v1)) 
   {
    *e00=NULL;
    *e01=NULL;
   }
   else scan_chain(e->edges[0],v1,e,v1,e00,e01);
   v1=e->vertices[1];
   if (IS_DANGLING_VERTEX(s,v1)) 
   {
    *e10=NULL;
    *e11=NULL;
   }
   else scan_chain(e->edges[1],v1,e,v1,e11,e10);
  }
  else                           /* regular case */
  {
   scan_chain(e->edges[0],e->vertices[0],e,e->vertices[1],e00,e10);
   scan_chain(e->edges[1],e->vertices[1],e,e->vertices[0],e11,e01);
  }
  return(TRUE);
 }
}

bool EDGE_VERTEX(s,e,v0,v1)
sptr s; eptr e; vptr *v0,*v1;
{
 if (! IS_AN_EDGE(s,e))
 {
  *v0=NULL;
  *v1=NULL;
  return(FALSE);
 }
 else
 {
  *v0=e->vertices[0];
  *v1=e->vertices[1];
  return(TRUE);
 }
}

bool EDGE_FACE(s,e,f0,f1)
sptr s; eptr e; fptr *f0,*f1;
{
 if (! IS_AN_EDGE(s,e))
 {
  *f0=NULL;
  *f1=NULL;
  return(FALSE);
 }
 else
 {
  *f0=e->faces[0];
  *f1=e->faces[1];
  return(TRUE);
 }
}

lptr FACE_EDGE(s,f)
sptr s; fptr f;
{
 eptr e,e1; vptr v,v1; lptr el;

 if (IS_A_FACE(s,f))
 {
  e=f->boundingedge;
  INIT(&el);
  if (e!=NULL)
  {
   v=e->vertices[f==e->faces[0]];
   e1=e;
   v1=v;
   do
   {
/*  if (e1->faces[0]!=e1->faces[1]) */
    INSERT_CLOCK(&el,e1);
    v1=e1->vertices[v1==e1->vertices[0]];
    e1=e1->edges[v1==e1->vertices[1]];
   } while (e1!=e || v1!=v);
  }
 }
 return(el);
}

lptr FACE_VERTEX(s,f)
sptr s; fptr f;
{
 eptr e,e1; vptr v,v1; lptr vl;

 INIT(&vl);
 if (IS_A_FACE(s,f))
 {
  e=f->boundingedge;
  if (e!=NULL)
  {
   v=e->vertices[f==e->faces[0]];
   e1=e;
   v1=v;
   do
   {
/*  if (e1->faces[0]!=e1->faces[1] */
    INSERT_CLOCK(&vl,v1);
    v1=e1->vertices[v1==e1->vertices[0]];
    e1=e1->edges[v1==e1->vertices[1]];
   } while (e1!=e || v1!=v);
  }
 }
 return(vl);
}

lptr FACE_FACE(s,f)
sptr s; fptr f;
{
 eptr e,e1; vptr v,v1; fptr f1,lastf=NULL; lptr fl;

 INIT(&fl);
 if (IS_A_FACE(s,f))
 {
  e=f->boundingedge;
  if (e!=NULL)
  {
   v=e->vertices[f==e->faces[0]];
   e1=e;
   v1=v;
   do
   {
    f1=e1->faces[v1==e1->vertices[1]];
    if (f1!=f && f1!=lastf);
    {
     INSERT_CLOCK(&fl,f1);
     lastf=f1;
    }
    v1=e1->vertices[v1==e1->vertices[0]];
    e1=e1->edges[v1==e1->vertices[1]];
   } while (e1!=e || v1!=v);
  }
 }
 return(fl);
}

lptr VERTEX_EDGE(s,v)
sptr s; vptr v;
{
 eptr e,e1; lptr el;

 INIT(&el);
 if (IS_A_VERTEX(s,v))
 {
  e=v->incidentedge;
  if (e!=NULL)
  {
   e1=e;
   do
   {
    INSERT_COUNTERCLOCK(&el,e1);
    e1=e1->edges[v==e1->vertices[1]];
   } while (e1!=e);
  }
 }
 return(el);
}

lptr VERTEX_VERTEX(s,v)
sptr s; vptr v;
{
 eptr e,e1; lptr vl;

 INIT(&vl);
 if (IS_A_VERTEX(s,v))
 {
  e=v->incidentedge;
  if (e!=NULL)
  {
   e1=e;
   do
   {
    INSERT_COUNTERCLOCK(&vl,e1->vertices[v==e1->vertices[0]]);
    e1=e1->edges[v==e1->vertices[1]];
   } while (e1!=e);
  }
 }
 return(vl);
}

lptr VERTEX_FACE(s,v)
sptr s; vptr v;
{
 eptr e,e1; fptr f1; lptr fl;

 INIT(&fl);
 if (IS_A_VERTEX(s,v))
 {
  e=v->incidentedge;
  if (e!=NULL)
  {
   e1=e;
   do
   {
    f1=e1->faces[v==e1->vertices[0]];
/*  if (f1!=FIRST(el)) */
    INSERT_COUNTERCLOCK(&fl,f1);
    e1=e1->edges[v==e1->vertices[1]];
   } while (e1!=e);
  }
 }
 return(fl);
}


/* Partial relations */

bool EXTREME_VERTICES(s,e,v0,v1)
sptr s; eptr e; vptr *v0,*v1;
{
 if (! IS_AN_EDGE(s,e))
 {
  *v0=NULL;
  *v1=NULL;
  return(FALSE);
 }
 else
 {
  *v0=e->vertices[0];
  *v1=e->vertices[1];
  return(TRUE);
 }
}
                  
vptr COMMON_VERTEX(s,e,e1)
sptr s; eptr e,e1;
{
 if (IS_AN_EDGE(s,e) && IS_AN_EDGE(s,e1))
 {
  if (e->vertices[0]==e1->vertices[0] ||
      e->vertices[0]==e1->vertices[1])
   return(e->vertices[0]);
  else 
     if (e->vertices[1]==e1->vertices[0] ||
         e->vertices[1]==e1->vertices[1])
   return(e->vertices[1]);
     else
      return(NULL);
 }
 else
  return(NULL);
}

eptr COMMON_EDGE(s,v,v1)
sptr s; vptr v,v1;
{
 eptr e; lptr l,l1;

 if ((IS_A_VERTEX(s,v) && IS_A_VERTEX(s,v1)) &&
     (! (IS_ISOLATED(s,v) || IS_ISOLATED(s,v1))))
 {
  l=VERTEX_EDGE(s,v);
  e=(eptr)FIRST(l);
  l1=MARK_POSITION(l);
  CLOCKWISE(&l);
  e=(eptr)FIRST(l);
  for (; !(SAME_POSITION(l,l1) || IS_INCIDENT(s,e,v1));
       CLOCKWISE(&l),e=(eptr)FIRST(l));
  if (IS_INCIDENT(s,e,v1))
  {
   FREELIST(&l);
   return(e);
  }
  else
  {
   FREELIST(&l);
   return(NULL);
  }
 }
 else 
 {
  FREELIST(&l);
  return(NULL);
 }
}

fptr OTHER_FACE(s,f,e)
sptr s; fptr f; eptr e;
{
 if (IS_AN_EDGE(s,e))
 {
  if (e->faces[0]==f)
   return(e->faces[1]);
  else if (e->faces[1]==f)
   return(e->faces[0]);
  else return(NULL);
 } 
 else return(NULL);
}
 
vptr OTHER_VERTEX(s,v,e)
sptr s; vptr v; eptr e;
{
 if (IS_AN_EDGE(s,e))
 {
  if (e->vertices[0]==v)
   return(e->vertices[1]);
  else
  {
   if (e->vertices[1]==v)
    return(e->vertices[0]);
   else return(NULL);
  }
 } 
 else return(NULL);
}

eptr FIRST_BOUNDING_EDGE(s,f)
sptr s; fptr f;
{
 if (IS_A_FACE(s,f))
  return(f->boundingedge);
 else
  return(NULL);
}

eptr NEXT_CLOCK_EDGE(s,f,e)
sptr s; fptr f; eptr e;
{
 if (IS_BOUNDING(s,e,f))
  return(e->edges[f==e->faces[1]]);
 else return(NULL);
} 

eptr CLOCK_INCIDENT_EDGE(s,f,v)
sptr s; fptr f; vptr v;
{
 lptr l; eptr e,e1; bool ok=FALSE;

 if (IS_A_VERTEX(s,v) && IS_A_FACE(s,f) && (! IS_ISOLATED(s,v)))
 {
  l=VERTEX_EDGE(s,v);
  e=(eptr)FIRST(l);
  e1=e;
  do
  {
   if (f==e1->faces[v==e1->vertices[0]] && f!=e1->faces[v==e1->vertices[1]])
    ok=TRUE;
   else
   {
    CLOCKWISE(&l);
    e1=(eptr)FIRST(l);
   }
  } while (!(ok || e1==e));
  if (ok) 
  {
   e=(eptr)FIRST(l);
   FREELIST(&l);
   return(e);
/*return((eptr)FIRST(l));*/
  }
  else 
  {
   FREELIST(&l);
   return(NULL);
  }
 }
 else
 {
  FREELIST(&l);
  return(NULL);
 }
}

vptr OPPOSITE_VERTEX(s,f,e)
sptr s; fptr f; eptr e;
{
 eptr e1;

 if (IS_A_TRIANGLE(s,f) && IS_BOUNDING(s,e,f))
 {
  e1=e->edges[e->faces[1]==f];
  return(OTHER_VERTEX(s,COMMON_VERTEX(s,e,e1),e1));
 }
 else return(NULL);
}
